import json
import os


def load_auth():
    """
    Load the authentication from the .slate.json file or the environment variables of the instance
    """
    try:
        slate_config = open('./slate.json')
        config = json.loads(slate_config.read())

        return config['model_id'], config['api_key'], config['api_pass']
    except FileNotFoundError:
        def validate_env(env, name):
            if env is None:
                raise EnvironmentError(f"A slate.json was not found and was not able to find a definition for the "
                                       f"environment variable: {name}")

        model = os.getenv('SLATE_MODEL_ID')
        validate_env(model, 'SLATE_MODEL_ID')

        api_key = os.getenv('SLATE_API_KEY')
        validate_env(api_key, 'SLATE_API_KEY')

        api_pass = os.getenv('SLATE_API_PASS')
        validate_env(api_pass, 'SLATE_API_PASS')

        return model, api_key, api_pass


def assemble_base(base_route: str, route: str) -> str:
    """
    Simple utility to append two route strings

    :param base_route: Base route like events.blankly.finance
    :param route: Specific route like
    :return: Appended strings as routes

    TODO this can validate that it is a valid path
    """
    return base_route + route



from slate.integrations.backtesting_py import BacktestingPy
from slate.integrations.bt_py import BtPy
from slate.integrations.jesse_ai import JesseAi


class Integrations:
    def __init__(self, slate, api):
        self.bt = BtPy(slate, api)
        self.backtesting = BacktestingPy(slate, api)
        self.jesse = JesseAi(slate, api)



from operator import itemgetter

import pandas as pd

import slate
from slate.integrations.common import b_id, DUMMY_METRICS, DUMMY_INDICATORS

try:
    import backtesting
except ImportError:
    pass


class BacktestingPy:
    slate: 'slate.Slate'
    api: 'slate.api.API'

    def __init__(self, slate, api):
        self.slate = slate
        self.api = api

    def post_backtest(self, result: 'backtesting.Backtest', symbol: str = None):
        self.slate.model.add_symbol(symbol)
        symbol = symbol or 'Unknown'
        quote = symbol.split('-')[1] if '-' in symbol else 'USD'

        trades = result['_trades'] \
            .apply(map_trades, axis=1, result_type='expand', symbol=symbol) \
            .unstack() \
            .reset_index(drop=True) \
            .apply(pd.Series) \
            .sort_values('time', ascending=True) \
            .to_dict('records')

        equity = result['_equity_curve']['Equity']
        account_values = equity.loc[equity.shift() != equity] \
            .reset_index() \
            .rename(columns={'index': 'time', 'Equity': 'value'})
        account_values['time'] = account_values['time'].map(lambda t: t.timestamp())
        account_values = account_values.to_dict('records')

        id = b_id()
        self.slate.backtest.result(symbols=[symbol],
                                   quote_asset=quote,
                                   start_time=result['Start'].timestamp(),
                                   stop_time=result['End'].timestamp(),
                                   account_values=account_values,
                                   trades=trades,
                                   backtest_id=id,
                                   metrics=DUMMY_INDICATORS,
                                   indicators=DUMMY_METRICS)

        self.slate.backtest.status(backtest_id=id,
                                   successful=True,
                                   status_summary='Completed',
                                   status_details='',
                                   time_elapsed=0)


def map_trades(row: pd.Series, symbol: str) -> list:
    common = {'symbol': symbol,
              'size': abs(row['Size']),
              'type': 'market'}
    entry = {**common,
             'time': row['EntryTime'].timestamp(),
             'side': 'buy' if row['Size'] > 0 else 'sell',
             'id': b_id(),
             'price': row['EntryPrice']}
    exit = {**common,
            'time': row['ExitTime'].timestamp(),
            'side': 'sell' if row['Size'] > 0 else 'buy',
            'id': b_id(),
            'price': row['ExitPrice']}
    return [entry, exit]


if __name__ == '__main__':
    # run backtesting.py backtest
    from backtesting import Backtest, Strategy
    from backtesting.lib import crossover
    from backtesting.test import SMA, GOOG


    class SmaCross(Strategy):
        n1 = 10
        n2 = 20

        def init(self):
            close = self.data.Close
            self.sma1 = self.I(SMA, close, self.n1)
            self.sma2 = self.I(SMA, close, self.n2)

        def next(self):
            if crossover(self.sma1, self.sma2):
                self.buy()
            elif crossover(self.sma2, self.sma1):
                self.sell()


    bt = Backtest(GOOG, SmaCross,
                  cash=10000, commission=.002,
                  exclusive_orders=True)
    result = bt.run()

    # post to slate
    slate = slate.Slate()
    slate.integrations.backtesting.post_backtest(result, 'GOOG')



import asyncio
import time
import typing

import slate.utils as utils
from slate.api import API

from slate.integrations import Integrations
from slate.live.live import Live
from slate.model.model import Model
from slate.backtest.backtest import Backtest


class Slate:
    def __init__(self, model_id: str = None, enable_async=False):
        """
        Initialize a new slate instance

        :param enable_async: Enable this to allow submission to the event loop
        """
        self.model_id, self.__api_key, self.__api_pass = utils.load_auth()
        if model_id is not None:
            self.model_id = model_id

        self.__api = API(self.model_id, self.__api_key, self.__api_pass)

        self.live = Live(self.__api)
        self.model = Model(self.__api)
        self.backtest = Backtest(self.__api)
        self.integrations = Integrations(self, self.__api)

        self.__settings = {
            'enable_async': enable_async
        }
        self.__event_loop = None
        if enable_async:
            self.__event_loop = asyncio.get_event_loop()
            self.__event_loop.run_forever()

    def submit(self, callable_: typing.Callable, **kwargs):
        """
        Submit a new job into the slate-managed event loop. Ensure that the enable_async is == True on the
         initialization

        :param callable_: The function to submit to the event loop
        :param kwargs: A key/value set of arguments for the function to evaluate
        :return:
        """
        if self.__event_loop is not None:
            self.__event_loop.create_task(self.__execute(callable_, kwargs))
        else:
            raise Exception("Must set enable_sync to True to submit to the event loop")

    @staticmethod
    async def __execute(callable_: typing.Callable, kwargs: dict):
        """
        Execute the function in the event loop

        :param callable_: The function to submit to the event loop
        :param kwargs: A key/value set of arguments for the function to evaluate
        :return:
        """
        return callable_(**kwargs)

    @property
    def now(self):
        return time.time()



import datetime
import json
import warnings

from slate.api import API
from slate.utils import assemble_base


class Live:
    def __init__(self, api: API):
        """
        Initialize a new live instance
        :param api: The object containing the Blankly API
        """
        self.__api: API = api

        self.__live_base = '/v1/live'

    def __assemble_base(self, route: str) -> str:
        """
        Assemble the sub-route specific to live posts

        :param route: The sub route: /spot-market -> /v1/live/spot-market
        :return: str
        """
        return assemble_base(self.__live_base, route)

    def event(self, args: dict, response: dict, type_: str, annotation: str = None,
              time: datetime.datetime = None) -> dict:
        """
        Post an event to the platform - generally used for annotating & viewing any important custom function calls
        https://docs.blankly.finance/services/events/#post-v1liveupdate-trade

        :param args: The (function) arguments
        :param response: The (function) response
        :param type_: A custom type for the event like 'order' or 'check_price'
        :param annotation: A human-friendly bit of text for your function
        :param time: A time object to fill if the event occurred in the past
        :return: API Response
        """
        return self.__api.post(self.__assemble_base('/event'), {
            'args': args,
            'response': response,
            'type': type_,
            'annotation': annotation
        }, time)

    def set_pnl(self, pnl_values: list) -> dict:
        """
        Submit a set of PNL values to the platform. This will overwrite any existing PNL

        :param pnl_values: An array of pnl values formatted [{time: 1651647605, value: 10},
         {time: 1651647605, value: 10}]
         :return: API Response
        """
        return self.__api.post(self.__assemble_base('/set-pnl'), {
            'values': json.dumps(pnl_values)
        })

    def set_auto_pnl(self, setting: bool) -> dict:
        """
        Enable or disable blankly's auto PNL on your trades

        :param setting: True to enable auto PNL, False to disable
        :return: API response
        """
        return self.__api.post(self.__assemble_base('/auto-pnl'), {
            'enabled': setting
        })

    def spot_market(self,
                    symbol: str,
                    exchange: str,
                    id_: str,
                    side: str,
                    size: [float, int] = None,
                    funds: [float, int] = None,
                    annotation: str = None,
                    time: datetime.datetime = None) -> dict:
        """
        Post a market order done on a spot exchange to the platform
        https://docs.blankly.finance/services/events/#post-v1liveupdate-trade

        :param symbol: The symbol the order is under
        :param exchange: The exchange the order was performed on - ex: 'coinbase_pro' or 'alpaca'
        :param id_: The exchange-given order id
        :param side: The side - 'buy' or 'sell'
        :param size: Mutually exclusive with funds - if you place an order priced in the base asset
        :param funds: Mutually exclusive with size - if you place an order priced in the quote asset
        :param annotation: An optional annotation to be given to this order
        :param time: A time object to fill if the event occurred in the past
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/spot-market'), {
            'symbol': symbol,
            'id': id_,
            'side': side,
            'exchange': exchange,
            'size': size,
            'funds': funds,
            'annotation': annotation
        }, time)

    def spot_limit(self,
                   symbol: str,
                   exchange: str,
                   id_: str,
                   side: str,
                   price: [int, float],
                   size: [float, int] = None,
                   funds: [float, int] = None,
                   annotation: str = None,
                   time: datetime.datetime = None):
        """
        Post a limit order done on a spot exchange to the platform
        https://docs.blankly.finance/services/events/#post-v1liveupdate-trade

        :param symbol: The symbol the order was placed on
        :param exchange: The exchange the order was performed on - ex: 'coinbase_pro' or 'alpaca'
        :param id_: The exchange-generated order id
        :param side: The order side - 'buy' or 'sell'
        :param price: The limit order price
        :param size: Mutually exclusive with funds - if you place an order priced in the base asset
        :param funds: Mutually exclusive with size - if you place an order priced in the quote asset
        :param annotation: An optional annotation to be given to this order
        :param time: A time object to fill if the event occurred in the past
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/spot-limit'), {
            'symbol': symbol,
            'id': id_,
            'side': side,
            'exchange': exchange,
            'price': price,
            'size': size,
            'funds': funds,
            'annotation': annotation
        }, time)

    def spot_stop(self,
                  symbol: str,
                  exchange: str,
                  id_: str,
                  side: str,
                  price: [int, float],
                  activate: [int, float],
                  size: [float, int] = None,
                  funds: [float, int] = None,
                  annotation: str = None,
                  time: datetime.datetime = None):
        """
        Post a limit order done on a spot exchange to the platform
        https://docs.blankly.finance/services/events/#post-v1liveupdate-trade

        :param symbol: The symbol the order was placed on
        :param exchange: The exchange the order was performed on - ex: 'coinbase_pro' or 'alpaca'
        :param id_: The exchange-generated order id
        :param side: The order side - 'buy' or 'sell'
        :param price: The limit order price
        :param activate: The activation price of the stop order
        :param size: Mutually exclusive with funds - if you place an order priced in the base asset
        :param funds: Mutually exclusive with size - if you place an order priced in the quote asset
        :param annotation: An optional annotation to be given to this order
        :param time: A time object to fill if the event occurred in the past
        :return: API response (dict)
        """
        return self.__api.post(self.__assemble_base('/spot-limit'), {
            'symbol': symbol,
            'id': id_,
            'side': side,
            'exchange': exchange,
            'price': price,
            'activate': activate,
            'size': size,
            'funds': funds,
            'annotation': annotation,
        }, time)

    def update_trade(self, id_: str, **kwargs) -> dict:
        """
        Update an existing trade on the platform by order ID
        https://docs.blankly.finance/services/events/#post-v1liveupdate-trade

        :param id_: The exchange-given order id
        :param kwargs: Any key/value pair. Generally these should be the same as the above order keys
        :return: API Response (dict)
        """
        kwargs['id'] = id_
        return self.__api.post(self.__assemble_base('/update-trade'), kwargs)

    def update_annotation(self, id_: str, annotation: str) -> dict:
        """
        Explicitly update just the order annotation
        https://docs.blankly.finance/services/events/#post-v1liveupdate-annotation

        :param id_: The exchange-given order id
        :param annotation: A descriptor about the order
        :return: API Response (dict)
        """
        return self.__api.post(self.__assemble_base('/update-annotation'), {
            'id': id_,
            'annotation': annotation
        })

    def screener_result(self, result: dict, time_: datetime.datetime = None) -> dict:
        """
        Post a screener result to the platform
        https://docs.blankly.finance/services/events/#post-v1livescreener-result

        :param result: The screener results object
            example:
            {
                "AAPL": {              // Organize by symbol
                    "RSI": 29.5,       // Keep the same keys internally across all symbols
                    "buy_signal": true // Keep same keys
                },
                "MSFT": {
                    "RSI": 54.3,       // Nesting beyond this level is not supported
                    "buy_signal": false
                }
            }
        :param time_: A time object to fill if the event occurred in the past
        :return: API response (dict)
        """

        # Test for double nesting in the results dictionary
        for i in result:
            inner = result[i]
            for j in inner:
                if isinstance(inner[j], dict):
                    warnings.warn(f"Double nested dictionaries are not supported in the screener result: "
                                  f"\n{json.dumps(inner[j], indent=2)}")

        # Format to API spec
        result = {'result': result}

        return self.__api.post(self.__assemble_base('/screener-result'), result, time_)

    def log(self, line: str, type_: str, time_: datetime.datetime = None) -> dict:
        """
        Post a new log from a live running model. This can be stdout or any custom logs
        https://docs.blankly.finance/services/events#post-v1livelog

        :param line: The line to write
        :param type_: The type of line, common types include 'stdout' or 'stderr'
        :param time_: A time object to fill if the event occurred in the past
        :return: API response (dict)
        """
        return self.__api.post('/log', {
            'line': line,
            'type': type_
        }, time_)
